--[[ =================================================================
    Description:
        Type /ubi or /uberinventory to see what you currently own,
        whether it is in the bank or in the bags you carry.
        So now where ever you are in the World of Warcraft you can
        get an overview of all your loot.

    Download:
        UberInventory - http://redeye.xs4all.nl/ubi/

    Contact:
        For questions, bug reports visit the website or send an email
        to the following address: wowaddon@xs4all.nl

    Dependencies:
        None

    Credits:
        A big 'Thank You' to all the people at Blizzard Entertainment
        for making World of Warcraft.
    ================================================================= --]]

-- Local globals
    local _G = _G;
    local strfind, strgsub, strlower = string.find, string.gsub, string.lower;
    local floor, ceil, abs, tonumber = math.floor, math.ceil, math.abs, tonumber;
    local pairs, tinsert, tsort, wipe = pairs, table.insert, table.sort, wipe;

-- Text and tooltips for checkbuttons
    UBI_CheckButtons = { { text = UBI_OPTION_MONEY, tooltip = UBI_OPTION_MONEY_TIP },
                         { text = UBI_OPTION_BALANCE, tooltip = UBI_OPTION_BALANCE_TIP },
                         { text = UBI_OPTION_SELLPRICES, tooltip = UBI_OPTION_SELLPRICES_TIP },
                         { text = UBI_OPTION_RECIPEPRICES, tooltip = UBI_OPTION_RECIPEPRICES_TIP },
                         { text = UBI_OPTION_QUESTREWARD, tooltip = UBI_OPTION_QUESTREWARD_TIP },
                         { text = UBI_OPTION_RECIPEDROP, tooltip = UBI_OPTION_RECIPEDROP_TIP },
                         { text = UBI_OPTION_TAKEMONEY, tooltip = UBI_OPTION_TAKEMONEY_TIP },
                         { text = UBI_OPTION_SHOWMAP, tooltip = UBI_OPTION_SHOWMAP_TIP },
                         { text = UBI_OPTION_GBSEND, tooltip = UBI_OPTION_GBSEND_TIP },
                         { text = UBI_OPTION_GBRECEIVE, tooltip = UBI_OPTION_GBRECEIVE_TIP },
                         { text = UBI_OPTION_WARN_MAILEXPIRE, tooltip = UBI_OPTION_WARN_MAILEXPIRE_TIP },
                         { text = UBI_OPTION_WARN_FULLMAILBOX, tooltip = UBI_OPTION_WARN_FULLMAILBOX_TIP },
                         { text = UBI_OPTION_HIGHLIGHT, tooltip = UBI_OPTION_HIGHLIGHT_TIP },
                       };

-- Static dialogs
    StaticPopupDialogs["UBI_CONFIRM_DELETE"] = {
        text = "%s",
        button1 = YES,
        button2 = NO,
        timeout = 10,
        whileDead = 1,
        exclusive = 1,
        showAlert = 1,
        hideOnEscape = 1
    };

-- UberInventory task: Collect mailbox cash
    function UBI_Task_CollectCash( mailid )
        if UBI_MAILBOX_OPEN then
            -- Mail info
            local _, _, mailSender, mailSubject, cashAttached = GetInboxHeaderInfo( mailid );

            if ( cashAttached > 0 ) then
                -- Report cash attached to message (from 3.0.2/3 duplicates seems to occur less)
                UberInventory_Message( UBI_MAIL_CASH:format( UberInventory_GetMoneyString( cashAttached ), mailSender, mailSubject ) );

                -- Take the money
                TakeInboxMoney( mailid );
            end;
        end;
    end;

-- UberInventory task: Turn UBI_COLLECTING_CASH and traverse mailbox for final update
    function UBI_Task_DoneCollectingCash( collecting )
        if ( collecting ) then
            -- Not collecting cash anymore
            UBI_COLLECTING_CASH = false;

            -- Traverse mailbox
            TaskHandlerLib:AddTask( "UberInventory", UberInventory_Save_Mailbox );
        end;
    end;

-- UberInventory task: Send stream start indication
    function UBI_Task_SendStart( version )
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:GBSTART "..version, "GUILD" );
        end;
    end;

-- UberInventory task: Collect mailbox cash
    function UBI_Task_SendData( data )
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:GBITEM "..data, "GUILD" );
        end;
    end;

-- UberInventory task: Send guildbank cash
    function UBI_Task_SendCash( cash )
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:GBCASH "..cash, "GUILD" );
        end;
    end;

-- UberInventory task: Send stream end indication
    function UBI_Task_SendEnd()
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:GBEND", "GUILD" );
        end;
    end;

-- Upgrade data structures
    function UberInventory_Upgrade()
        -- From global to local
        local UBI_Options = UBI_Options;
        local UBI_Data = UBI_Data;

        -- Upgrade process
        if ( UBI_Options["version"] ~= UBI_VERSION ) then
            -- Show upgrade message
            UberInventory_Message( UBI_UPGRADE );

            -- Upgrade to version 1.0
            if ( UBI_Options["version"] < "1.0" ) then
                if ( UBI_Data["ItemPrices"] ) then
                    UBI_Data["ItemPrices"] = nil;
                end;
            end;

            -- Upgrade to version 1.3
            if ( UBI_Options["version"] < "1.3" ) then
                if ( not UBI_Options["show_recipe_prices"] ) then
                    UBI_Options["show_recipe_prices"] = true;
                end;
                if ( not UBI_Options["show_recipe_reward"] ) then
                    UBI_Options["show_recipe_reward"] = true;
                end;
                if ( not UBI_Options["show_recipe_drop"] ) then
                    UBI_Options["show_recipe_drop"] = true;
                end;
            end;

            -- Upgrade to version 1.6
            if ( UBI_Options["version"] < "1.6" ) then
                if ( not UBI_Options["send_gb_data"] ) then
                    UBI_Options["send_gb_data"] = true;
                end;
                if ( not UBI_Options["receive_gb_data"] ) then
                    UBI_Options["receive_gb_data"] = true;
                end;
                if ( not UBI_Options["warn_mailexpire"] ) then
                    UBI_Options["warn_mailexpire"] = true;
                end;
                if ( not UBI_Options["warn_fullmailbox"] ) then
                    UBI_Options["warn_fullmailbox"] = true;
                end;
            end;

            -- Upgrade to version 1.7
            if ( UBI_Options["version"] < "1.7" ) then
                for key, value in pairs( UBI_Data[UBI_REALM] ) do
                    if ( key ~= "Guildbank" ) then
                        -- Upgrade character data
                        for key, value in pairs( value["Items"] ) do
                            value["itemid"] = tonumber( value["itemid"] );
                        end;
                    else
                        -- Upgrade guildbank data
                        for key, value in pairs( value ) do
                            for key, value in pairs( value["Items"] ) do
                                value["itemid"] = tonumber( value["itemid"] );
                            end;
                        end;
                    end;
                end;
            end;

            -- Upgrade to version 2.0
            if ( UBI_Options["version"] < "2.0" ) then
                if ( not UBI_Options["show_higlight"] ) then
                    UBI_Options["show_higlight"] = true;
                end;
            end;

            -- Set current version
            UBI_Options["version"] = UBI_VERSION;
        end;
    end;

-- Days since visit
    function UberInventory_DaysSince( visit )
        return floor( difftime( time(), visit or time() ) / UBI_SEC_IN_DAY );
    end;

-- Send message to the default chat frame
    function UberInventory_Message( msg, prefix )
        -- Initialize
        local prefixText = "";

        -- Add application prefix
        if ( prefix or true ) then
            prefixText = C_GREEN.."UBI: "..C_CLOSE;
        end;

        -- Send message to chatframe
        DEFAULT_CHAT_FRAME:AddMessage( prefixText..( msg or "" ) );
    end;

-- Split money (gold, silver, copper)
    function UberInventory_SplitMoney( amount )
        local gold = floor( amount / 10000 );
        local silver = floor( ( amount - ( gold * 10000 ) ) / 100 );
        local copper = floor( ( amount - ( gold * 10000 ) ) - ( silver * 100 ) );

        return gold, silver, copper;
    end;

-- Get money split for cash owned ( gold, silver, copper )
    function UberInventory_GetMoney()
        -- Save money
        UBI_Money["Cash"] = GetMoney();

        return UberInventory_SplitMoney( UBI_Money["Cash"] );
    end;

-- Build moneystring
    function UberInventory_GetMoneyString( money )
        -- Initialize
        local gold, silver, copper = UberInventory_SplitMoney( money );
        local moneystring = "";

        -- Build money string
        if ( gold > 0 ) then moneystring = moneystring.." "..gold.." "..UBI_GOLD_ICON; end;
        if ( silver > 0 ) then moneystring = moneystring.." "..silver.." "..UBI_SILVER_ICON; end;
        if ( copper > 0 or money == 0 ) then moneystring = moneystring.." "..copper.." "..UBI_COPPER_ICON; end;

        return moneystring:trim();
    end;

-- Specialized SetTooltipMoney function (till internal Blizzard code for SetTooltipMoney is fixed)
    function UberInventory_SetTooltipMoney( tooltip, money, type, prefix, suffix )
        -- Initialize
        local moneystring = UberInventory_GetMoneyString( money );

        -- Build the string
        if ( prefix ) then moneystring = prefix..moneystring; end;
        if ( suffix ) then moneystring = moneystring..suffix; end;

        -- Add info to the tooltip
        tooltip:AddLine( moneystring.."  ", 1.0, 1.0, 1.0 );
    end;

-- Update slot count information
    function UberInventory_Update_SlotCount()
        UberInventoryFrameBagSlots:SetFormattedText( UBI_SLOT_BAGS, UBI_Options["bag_free"] , UBI_Options["bag_max"] );
        UberInventoryFrameBankSlots:SetFormattedText( UBI_SLOT_BANK, UBI_Options["bank_free"], UBI_Options["bank_max"] );
    end;

-- Fetch item prices (buy, sell)
    function UberInventory_GetItemPrices( itemid )
        -- Initialize
        local buyPrice, sellPrice = UBI_Prices_Buy[itemid], UBI_Prices_Sell[itemid];

        return buyPrice, sellPrice;
    end;

-- Send verison information
    function UberInventory_SendVersion()
        -- Send version if in a guild
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:VERSION "..UBI_VERSION, "GUILD" );
        end;
    end;

-- Extract info from tooltip
    function UberInventory_Scan_Tooltip( itemid, color, searchtext )
        -- Rounding function
        local function round( x, digit )
            -- Positions to be shifted
            local shift = 10 ^ digit;

            -- Round the number to wanted decimals
            return floor( x * shift + 0.5 ) / shift;
        end;

        -- Initialize
        local txtLeft = "";
        local txtRight = "";
        local digits = 4;
        local colorFound = false;
        local textFound = false;
        local r, g, b, a, txtLeft, txtRight;

        -- Setup tooltip for extracting data
        local tooltip = UberInventory_ItemTooltip;
        tooltip:ClearLines();
        tooltip:SetHyperlink( UberInventory_GetLink( itemid ) );
        tooltip:Show();

        -- Correct color precision
        color.r = round( color.r, digits );
        color.g = round( color.g, digits );
        color.b = round( color.b, digits );

        -- Traverse tooltip lines
        for i = 1, tooltip:NumLines() do
            -- Get text
            txtLeft = _G[tooltip:GetName().."TextLeft"..i]:GetText();
            txtRight = _G[tooltip:GetName().."TextRight"..i]:GetText();

            -- Find the specified color
            if ( color ) then
                -- Check left color
                r, g, b, a = _G[tooltip:GetName().."TextLeft"..i]:GetTextColor();

                -- Check if requested color was used
                if ( ( round( r, digits ) == color.r ) and
                     ( round( g, digits ) == color.g ) and
                     ( round( b, digits ) == color.b ) and
                     txtLeft ) then
                    colorFound = true;
                end;

                -- Check right color
                local r, g, b, a = _G[tooltip:GetName().."TextRight"..i]:GetTextColor();

                -- Check if requested color was used
                if ( ( round( r, digits ) == color.r ) and
                     ( round( g, digits ) == color.g ) and
                     ( round( b, digits ) == color.b ) and
                     txtRight ) then
                    colorFound = true;
                end;
            end;

            -- Find the specified text
            if ( searchtext and txtLeft ) then
                textFound = txtLeft:match( searchtext );
            end;
            if ( searchtext and txtRight ) then
                textFound = txtRight:match( searchtext );
            end;

            -- Stop on first blank line or new-line char
            if ( txtLeft == "" or txtLeft:find( "\n" ) ) then
                return colorFound, textFound;
            end;
        end;

        return colorFound, textFound;
    end;

-- Can item used by current player
    function UberInventory_UsableItem( itemid )
        -- Initialize
        local TOOLTIP_RED = { r = 0.99999779462814, g = 0.12548992037773, b = 0.12548992037773 };

        -- Determine if the color red is used within the tooltip
        local colorFound, textFound = UberInventory_Scan_Tooltip( itemid, TOOLTIP_RED, ITEM_SPELL_KNOWN );

        -- Return result (usable or not)
        return ( ( not colorFound ) or textFound );
    end;

-- Build list of usable alts
    function UberInventory_GetAlts()
        -- Initialize
        local UBI_Characters = UBI_Characters;
        wipe( UBI_Characters );

        -- Traverse all players to determine valid alts
        for key, value in pairs( UBI_Data[UBI_REALM] ) do
            if ( key ~= "Guildbank" ) and
               ( key ~= UBI_PLAYER ) and
               ( ( UBI_Data[UBI_REALM][key]["Options"]["faction"] or "<Unknown>" ) == UBI_FACTION ) then
                -- Add alt to the list
                tinsert( UBI_Characters, key );
            end;
        end;

        -- Sort alts alphabetically
        tsort( UBI_Characters );
    end;

-- Build list of guildbanks
    function UberInventory_GetGuildbanks()
        -- Initialize
        local UBI_Guildbanks = UBI_Guildbanks;
        wipe( UBI_Guildbanks );

        -- Traverse all guildbanks
        for key, value in pairs( UBI_Guildbank ) do
            if ( ( value["Faction"] or "<Unknown>" ) == UBI_FACTION ) then
                -- Add alt to the list
                tinsert( UBI_Guildbanks, key );
            end;
        end;

        -- Sort alts alphabetically
        tsort( UBI_Guildbanks );
    end;

-- Populate locations dropdown box (Also contains items for alts on the same realm from the same faction (Alliance or Horde))
    function UberInventory_Locations_Init()
        -- Initialize
        local UBI_Item;
        local UBI_LocationCounter = UBI_LocationCounter;

        -- Clear locations and reset counter
        wipe( UBI_LocationList );
        UBI_LocationCounter = 0;

        -- Custom function for adding items to dropdownbox
        local function UberInventory_AddLocation( item, itemtype )
            UBI_LocationList[UBI_LocationCounter] = { key = item.value, name = item.text, value = strtrim( item.text ), type = itemtype };
            UBI_LocationCounter = UBI_LocationCounter + 1;

            -- Add item to dropdown
            UIDropDownMenu_AddButton( item );
        end;

        -- All items (inludes current characters, all guildbanks and all other characters)
        UBI_Item = { text = UBI_ALL_LOCATIONS,
                     value = UBI_LocationCounter,
                     owner = this:GetParent(),
                     notCheckable = 1,
                     func = UberInventory_Locations_OnClick,
                     colorCode = "|cFFFFFF00" };
        UberInventory_AddLocation( UBI_Item, "complete" );

        -- Insert divider
        UBI_Item = { text = "",
                     isTitle = 1,
                     owner = this:GetParent() };
        UberInventory_AddLocation( UBI_Item, "title" );

        -- Current characters
        UBI_Item = { text = UBI_PLAYER,
                     value = UBI_LocationCounter,
                     owner = this:GetParent(),
                     notCheckable = 1,
                     func = UberInventory_Locations_OnClick,
                     colorCode = "|cFFFFFF00" };
        UberInventory_AddLocation( UBI_Item, "current" );

        -- Add the basic locations (including Guildbank when member of a guild)
        for key, value in pairs( UBI_LOCATIONS ) do
            -- Set new item
            UBI_Item = { text = "  "..value,
                         value = UBI_LocationCounter,
                         icon = UBI_LOCATION_TEXTURE[key],
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Locations_OnClick };

            -- Add item to dropdown
            UberInventory_AddLocation( UBI_Item, "current" );
        end;

        -- Add separator if there are guildbanks to be added
        if ( #UBI_Guildbanks > 0 ) then
            -- Set new item
            UBI_Item = { text = "",
                         isTitle = 1,
                         owner = this:GetParent() };
            UberInventory_AddLocation( UBI_Item, "title" );

            UBI_Item = { text = UBI_ALL_GUILDBANKS,
                         value = UBI_LocationCounter,
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Locations_OnClick,
                         colorCode = "|cFFFFFF00" };
            UberInventory_AddLocation( UBI_Item, "all_guildbank" );
        end;

        -- Add guildbanks (how to determine faction for image?)
        for key, value in pairs( UBI_Guildbanks ) do
            -- Set new item
            UBI_Item = { text = "  "..value,
                         value = UBI_LocationCounter,
                         icon = UBI_LOCATION_TEXTURE[6],
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Locations_OnClick };

            -- Add highlight to current players guildbank
            if ( value == UBI_GUILD ) then
                UBI_Item.colorCode = ITEM_QUALITY_COLORS[5].hex;
            end;

            -- Add item to dropdown (Guildbank only added if in a guild)
            UberInventory_AddLocation( UBI_Item, "guildbank" );
        end;

        -- Add separator if there are alts to be added
        if ( #UBI_Characters > 0 ) then
            -- Set new item
            UBI_Item = { text = "",
                         isTitle = 1,
                         owner = this:GetParent() };
            UberInventory_AddLocation( UBI_Item, "title" );

            UBI_Item = { text = UBI_ALL_CHARACTERS,
                         value = UBI_LocationCounter,
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Locations_OnClick,
                         colorCode = "|cFFFFFF00" };
            UberInventory_AddLocation( UBI_Item, "all_character" );
        end;

        -- Add alts to the list from the same realm and faction (Alliance or Horde)
        for key, value in pairs( UBI_Characters ) do
            UBI_Item = { text = "  "..value,
                         value = UBI_LocationCounter,
                         icon = UBI_LOCATION_TEXTURE[6],
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Locations_OnClick };
            UberInventory_AddLocation( UBI_Item, "character" );
        end;
   end;

-- Update locations dropdown box
   function UberInventory_Locations_OnClick()
       UBI_FILTER_LOCATIONS = this.value;
       UIDropDownMenu_SetText( this.owner, UBI_LocationList[UBI_FILTER_LOCATIONS].value );
       UBI_PAGE = 0;
       UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
   end;

-- Populate quality dropdown box
    function UberInventory_Quality_Init()
        -- Initialize
        local UBI_Item;
        local UBI_QUALITY = UBI_QUALITY;

        for key, value in pairs( UBI_QUALITY ) do
            -- Set new item
            UBI_Item = { text = value,
                         value = key,
                         owner = this:GetParent(),
                         notCheckable = 1,
                         func = UberInventory_Quality_OnClick,
                         colorCode = ITEM_QUALITY_COLORS[key-2].hex };

            -- Correct color for 'All qualities'
            if (key-2 == -1) then
                UBI_Item.colorCode = ITEM_QUALITY_COLORS[1].hex;
            end;

            -- Add item to dropdown
            UIDropDownMenu_AddButton( UBI_Item );
        end;
    end;

-- Update quality dropdown box
   function UberInventory_Quality_OnClick()
       UBI_FILTER_QUALITY = this.value;
       UBI_PAGE = 0;
       UIDropDownMenu_SetText( this.owner, UBI_QUALITY[UBI_FILTER_QUALITY] );
       UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
   end;

-- Populate classes dropdown box
    function UberInventory_Classes_Init()
        -- Initialize
        local subclasses, UBI_Item;
        local UBI_CLASSES = UBI_CLASSES;

        -- Build structure for classes
        if ( UIDROPDOWNMENU_MENU_LEVEL == 1 ) then
            for key, value in pairs( UBI_CLASSES ) do
                -- Determine subclasses
                subclasses = { GetAuctionItemSubClasses( key-1 ) };

                -- Set new item
                UBI_Item = { text = value,
                             value = { key, 0 },
                             owner = UberInventoryFrameClassDropDown,
                             notCheckable = 1,
                             hasArrow = #subclasses > 0,
                             func = UberInventory_Classes_OnClick };

                -- Add item to dropdown
                UIDropDownMenu_AddButton( UBI_Item, UIDROPDOWNMENU_MENU_LEVEL );
            end;
        end;

        -- Build structure for subclasses
        if ( UIDROPDOWNMENU_MENU_LEVEL == 2 ) then
            -- Determine subclasses
            subclasses = { GetAuctionItemSubClasses( UIDROPDOWNMENU_MENU_VALUE[1]-1 ) };

            for key, value in pairs( subclasses ) do
                -- Set new item
                UBI_Item = { text = value,
                             value = { UIDROPDOWNMENU_MENU_VALUE[1], key },
                             owner = UberInventoryFrameClassDropDown,
                             notCheckable = 1,
                             func = UberInventory_Classes_OnClick
                             };

                -- Add item to dropdown
                UIDropDownMenu_AddButton( UBI_Item, UIDROPDOWNMENU_MENU_LEVEL );
            end;
        end;
    end;

-- Update classes dropdown box
    function UberInventory_Classes_OnClick()
        -- Get selection
        UBI_FILTER_CLASSES = this.value[1];
        UBI_FILTER_SUBCLASSES = this.value[2];

        -- Determine dropdown text
        local infoText = UBI_CLASSES[UBI_FILTER_CLASSES];
        if ( UBI_FILTER_SUBCLASSES > 0 ) then
            local subclasses = { GetAuctionItemSubClasses( UBI_FILTER_CLASSES-1 ) };
            infoText = infoText.." > "..subclasses[UBI_FILTER_SUBCLASSES];
            HideDropDownMenu( 1 ); -- Collapse also the top level menu
        end;

        UBI_PAGE = 0;
        UIDropDownMenu_SetText( this.owner, infoText );
        UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
    end;

-- OnLoad scroll frame
    function UberInventory_Scroll_OnLoad( scrollbar )
        UBI_SCROLLBAR = _G[scrollbar];
        UBI_SCROLLBAR:SetValueStep( 1 );
        UBI_SCROLLBAR.scrollBarHideable = true;
    end;

-- Toggle scroll buttons
    function UberInventory_ToggleButtons( pages )
        -- Enable/disable scroll buttons
        local min, max = UBI_SCROLLBAR:GetMinMaxValues();
        if ( min == UBI_SCROLLBAR:GetValue() ) then
            _G[UBI_SCROLLBAR:GetName().."ScrollUpButton"]:Disable();
        else
            _G[UBI_SCROLLBAR:GetName().."ScrollUpButton"]:Enable();
        end;
        if ( max == UBI_SCROLLBAR:GetValue() or pages < 0 ) then
            _G[UBI_SCROLLBAR:GetName().."ScrollDownButton"]:Disable();
        else
            _G[UBI_SCROLLBAR:GetName().."ScrollDownButton"]:Enable();
        end;
    end;

-- Initialize scroll frame
    function UberInventory_Scroll_Init( pages )
        -- Initiliaze some settings
        if ( pages or 0 ) < 0 then pages = 0 end;

        -- Set new scrollbar values
        UBI_SCROLLBAR:SetMinMaxValues( 0, pages * UBI_PIX_PER_PAGE );
        UBI_SCROLLBAR:SetValue( UBI_PAGE * UBI_PIX_PER_PAGE );

        -- Update scrollbuttons
        UberInventory_ToggleButtons( pages );
    end;

-- Handle click events from scroll frame
    function UberInventory_Scroll_Click()
        -- Calculate current page
        UBI_PAGE = ceil( UBI_SCROLLBAR:GetValue() / UBI_PIX_PER_PAGE );

        -- Update inventory frame
        UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

        -- Update scrollbuttons
        UberInventory_ToggleButtons( UBI_PAGE );
    end;

-- Returns an itemLink
    function UberInventory_GetLink( itemId )
        -- Get name of item
        local itemName, _, itemQuality = GetItemInfo( itemId );

        -- If the item has never been seen before, itemName will be nil
        if ( itemName == nil ) then
            itemName = UBI_ITEM_UNCACHED;
            itemQuality = 0;
        end;

        -- Determine item color
        local color = ITEM_QUALITY_COLORS[itemQuality].hex;

        -- Return itemLink
        return color.."|Hitem:"..itemId..":0:0:0:0:0:0:0|h["..itemName.."]|h|r";
    end;

-- Handle OnEnter event of the Minimap icon
    function UberInventory_Minimap_OnEnter( anchor )
        -- Set location of tooltip and clear old content
        GameTooltip:SetOwner( this, anchor );

        -- Add title (AddOn name and version)
        GameTooltip:AddDoubleLine( UBI_NAME, UBI_VERSION,
                                   HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b,
                                   NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b );

        -- Add slot information
        GameTooltip:AddDoubleLine( UBI_BAG, UBI_FREE:format( UBI_Options["bag_free"] , UBI_Options["bag_max"] ) );
        GameTooltip:AddDoubleLine( UBI_BANK, UBI_FREE:format( UBI_Options["bank_free"] , UBI_Options["bank_max"] ) );

        -- Add cash inforamtion
        GameTooltip:AddDoubleLine( UBI_CASH_TOTAL, UberInventory_GetMoneyString( GetMoney() + UBI_Money["mail"] ) );

        -- (Re)display the tooltip
        GameTooltip:Show();
    end;

-- Sorting table content
    function UberInventory_SortFilter( str, cmdline )
        -- From global to local
        local UBI_Track = UBI_Track;
        local UBI_Sorted = UBI_Sorted;
        local UBI_Items_Work = UBI_Items_Work;

        -- determine if sorting/collecting is needed
        if ( UBI_Filter.text == UBI_FILTER_TEXT and
             UBI_Filter.location == UBI_FILTER_LOCATIONS and
             UBI_Filter.quality == UBI_FILTER_QUALITY and
             UBI_Filter.class == UBI_FILTER_CLASSES and
             UBI_Filter.subclass == UBI_FILTER_SUBCLASSES and
             UBI_Filter.usable == UBI_FILTER_USABLE and
             UBI_Filter.date == UBI_SCAN_DATE ) and not UberInventoryFrame:IsShown() then
            return;
        else
            -- Store new filter settings
            UBI_Filter = { text = UBI_FILTER_TEXT,
                           location = UBI_FILTER_LOCATIONS,
                           quality = UBI_FILTER_QUALITY,
                           class = UBI_FILTER_CLASSES,
                           subclass = UBI_FILTER_SUBCLASSES,
                           usable = UBI_FILTER_USABLE,
                           date = UBI_SCAN_DATE };
        end;

        -- Initialize globals
        wipe( UBI_Sorted );
        wipe( UBI_Track );
        wipe( UBI_Items_Work );
        UBI_Inventory_count = 0;

        -- Copy items to separate array for sorting and displaying
        local include_item;
        local subclasses;
        local function UberInventory_CopyItems( str, cmdline )
            -- Copy items to local array (including filter)
            for key, record in pairs( UBI_Items_Work ) do
                -- Initialize
                include_item = true;

                -- Count number of unique items
                if ( not UBI_Track[record.itemid] ) then
                    UBI_Inventory_count = UBI_Inventory_count + 1;
                    UBI_Track[record.itemid] = 0;
                end;

                -- Fix totalCount (only for actual characters)
                if ( record.bag_count ) then
                    record.total = record.bag_count + record.bank_count + record.keyring_count +
                                   record.mailbox_count + record.equip_count;
                end;

                -- Only apply filters when searching from then inventory frame
                if ( not cmdline ) then
                    -- Filter on location (2 = Character, 3 = Bags, 4 = Banks, 5 = Keyring, 6 = Mail, 7 = Equipped )
                    if ( ( UBI_FILTER_LOCATIONS == 3 and record.bag_count == 0 ) or
                         ( UBI_FILTER_LOCATIONS == 4 and record.bank_count == 0 ) or
                         ( UBI_FILTER_LOCATIONS == 5 and record.keyring_count == 0 ) or
                         ( UBI_FILTER_LOCATIONS == 6 and record.mailbox_count == 0 ) or
                         ( UBI_FILTER_LOCATIONS == 7 and record.equip_count == 0 ) ) then
                        include_item = false;
                    end;

                    -- Filter on quality
                    if ( UBI_FILTER_QUALITY > 1 and record.quality ~= UBI_FILTER_QUALITY-2 ) then
                        include_item = false;
                    end;

                    -- Filter on class
                    if ( UBI_FILTER_CLASSES > 1 and record.type ~= UBI_CLASSES[UBI_FILTER_CLASSES] ) then
                        include_item = false;
                    end;

                    -- Filter on subclass
                    subclasses = { GetAuctionItemSubClasses( UBI_FILTER_CLASSES-1 ) };
                    if ( UBI_FILTER_SUBCLASSES > 0 and record.subtype ~= subclasses[UBI_FILTER_SUBCLASSES] ) then
                        include_item = false;
                    end;

                    -- Filter on usability
                    if ( UBI_FILTER_USABLE ) then
                        if ( not UberInventory_UsableItem( record.itemid ) ) then
                            include_item = false;
                        end;
                    end;
                end;

                -- Filter on name and store into structure suitable for sorting
                if ( ( strfind( strlower( record.name ) , strlower( str ) ) or 0 ) > 0 ) and include_item then
                    -- Insert item into sorting structure
                    if ( UBI_Track[record.itemid] == 0 ) then
                        tinsert( UBI_Sorted, record );
                    end;

                    -- Track correct item totals
                    UBI_Track[record.itemid] = UBI_Track[record.itemid] + record.total;
                end;
            end;
        end;

        -- Set base list of items to use
        if ( cmdline or UBI_LocationList[UBI_FILTER_LOCATIONS].type == "current" ) then
            -- Current character
            UBI_Items_Work = UBI_Items;
            UberInventory_CopyItems( str, cmdline );
        elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "guildbank" ) then
            -- Guildbank
            UBI_Items_Work = UBI_Guildbank[UBI_LocationList[UBI_FILTER_LOCATIONS].value]["Items"];
            UberInventory_CopyItems( str, cmdline );
        elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "character" ) then
            -- Alt characters
            UBI_Items_Work = UBI_Data[UBI_REALM][UBI_LocationList[UBI_FILTER_LOCATIONS].value]["Items"];
            UberInventory_CopyItems( str, cmdline );
        elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "complete" ) then
            -- Traverse characters
            for key, value in pairs( UBI_Characters ) do
                UBI_Items_Work = UBI_Data[UBI_REALM][value]["Items"];
                UberInventory_CopyItems( str, cmdline );
            end;

            -- Current character
            UBI_Items_Work = UBI_Items;
            UberInventory_CopyItems( str, cmdline );

            -- Traverse guildbanks
            for key, value in pairs( UBI_Guildbanks ) do
                UBI_Items_Work = UBI_Guildbank[value]["Items"];
                UberInventory_CopyItems( str, cmdline );
            end;
        elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_guildbank" ) then
            -- Traverse guildbanks
            for key, value in pairs( UBI_Guildbanks ) do
                UBI_Items_Work = UBI_Guildbank[value]["Items"];
                UberInventory_CopyItems( str, cmdline );
            end;
        elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_character" ) then
            -- Current character
            UBI_Items_Work = UBI_Items;
            UberInventory_CopyItems( str, cmdline );

            -- Traverse characters
            for key, value in pairs( UBI_Characters ) do
                UBI_Items_Work = UBI_Data[UBI_REALM][value]["Items"];
                UberInventory_CopyItems( str, cmdline );
            end;
        end;

        -- Sort the array alphabetically based on item name
        tsort( UBI_Sorted, function( rec1, rec2 )
                               return rec1.name < rec2.name
                           end  );
    end;

-- Reset filters
    function UberInventory_ResetFilters()
        -- Reset current page
        UBI_PAGE = 0;

        -- Reset dropdown for location
        UBI_FILTER_LOCATIONS = 2;
        UIDropDownMenu_SetText( UberInventoryFrameLocationDropDown, UBI_LocationList[UBI_FILTER_LOCATIONS].value );

        -- Reset dropdown for quality
        UBI_FILTER_QUALITY = 1;
        UIDropDownMenu_SetText( UberInventoryFrameQualityDropDown, UBI_QUALITY[UBI_FILTER_QUALITY] );

        -- Reset dropdown for class
        UBI_FILTER_CLASSES = 1;
        UBI_FILTER_SUBCLASSES = 0;
        UIDropDownMenu_SetText( UberInventoryFrameClassDropDown, UBI_CLASSES[UBI_FILTER_CLASSES] );

        -- Reset usable items
        UBI_FILTER_USABLE = false;
        UberInventoryFrameUsableItems:SetChecked( UBI_FILTER_USABLE );

        -- Reset search text (and refresh items displayed)
        if ( UBI_FILTER_TEXT == "" ) then
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
        else
            UberInventoryFrameItemSearch:SetText( "" );
        end;
    end;

-- Reset item counts
    function UberInventory_ResetCount( location )
        -- From global to local
        local UBI_Items = UBI_Items;

        -- Clear counts
        location = location.."_count";
        for key, record in pairs( UBI_Items ) do
            record[location] = 0;
        end;
    end;

-- Add sell prices to items in your bag (only used if not already at a merchant)
    function UberInventory_AddItemInfo( tooltip )
        -- Function for adding counts per location (only non-zero locations are added to the tooltip)
        local function AddLocationInfo( tooltip, bag, bank, keyring, mailbox, equipped )
            tooltip:AddLine( "|n" );
            if ( bag > 0 ) then tooltip:AddDoubleLine( UBI_LOCATIONS[1], bag ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[1] ); end;
            if ( bank > 0 ) then tooltip:AddDoubleLine( UBI_LOCATIONS[2], bank ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[2] ); end;
            if ( keyring > 0 ) then tooltip:AddDoubleLine( UBI_LOCATIONS[3], keyring ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[3] ); end;
            if ( mailbox > 0 ) then tooltip:AddDoubleLine( UBI_LOCATIONS[4], mailbox ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[4] ); end;
            if ( equipped > 0 ) then tooltip:AddDoubleLine( UBI_LOCATIONS[5], equipped ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[5] ); end;
        end;

        -- Set smaller font for vendor, quest and drop information
        local function SetSmallerFont()
            local numlines = tooltip:NumLines();
            local txtLeft = _G[ tooltip:GetName().."TextLeft"..numlines ];
            local txtRight = _G[ tooltip:GetName().."TextRight"..numlines ];
            txtLeft:SetFontObject( GameTooltipTextSmall );
            txtLeft:SetTextColor( 0.0, 1.0, 0.0 );
            txtRight:SetFontObject( GameTooltipTextSmall );
            txtRight:SetTextColor( 0.0, 0.7, 1.0 );
        end;

        -- Initialize
        local itemCount = 1;
        local guildCount = 0;
        local spacerAdded = false;

        -- Get the itemLink
        local itemName, itemLink = tooltip:GetItem();

        -- Does the slot contain an item
        if ( itemLink ) then
            -- Get item ID
            local _, itemId = strsplit( ":", itemLink );
            local _, _, _, _, _, itemType, itemSubType = GetItemInfo( itemLink );

            -- Get count (only if count label exists and is visible, otherwise use the value 1)
            if ( tooltip:GetName() == "ItemRefTooltip" ) then
                itemCount = 1;
            else
                -- Special fix needed for Auctioneer Advanced, since this AddOn removes itembutton as soon as
                -- the mouse is over it to display a bigger button on a different location
                if ( GetMouseFocus() and GetMouseFocus():GetName() ) then
                  local objCount = _G[ GetMouseFocus():GetName().."Count" ];
                  if ( objCount and objCount:IsVisible() ) then
                      itemCount = tonumber( objCount:GetText() ) or 1;
                  end;
                else
                    itemCount = 1;
                end;
            end;

            -- Get stored sell price
            local itemId_num = tonumber( itemId );
            local buyPrice, sellPrice = UberInventory_GetItemPrices( itemId_num );
            local name, area, mob, rate, count;
            local UBI_Zones = UBI_Zones;
            local totalCount = 0;

            -- Only perform the following part when it not has already been done (gets called twice for recipe related items)
            if ( not tooltip.UBI_InfoAdded or false ) then

                -- Add additional pricing information when not at a vendor
                if ( not MerchantFrame:IsVisible() or tooltip == ItemRefTooltip ) then
                    -- Add pricing information for selling to vendors
                    if ( UBI_Options["show_sell_prices"] ) then
                        if ( ( sellPrice or 0 ) == 0 ) then  -- Item unsellable
                            tooltip:AddLine( "|n" );
                            tooltip:AddLine( ITEM_UNSELLABLE, 1.0, 1.0, 1.0 );
                        elseif ( itemCount > 0 ) then -- Calculate sell price and add it
                            -- Add money to the tooltip
                            tooltip:AddLine( "|n" );
                            UberInventory_SetTooltipMoney( tooltip, sellPrice * itemCount, "STATIC", UBI_ITEM_SELL, "" );
                        end;
                    end;

                    -- Add pricing information for buying recipes from vendors (mainly for auction house use)
                    local recipeData = UBI_RecipeVendors[itemId_num];
                    if ( UBI_Options["show_recipe_prices"] and recipeData ) then
                        -- Add section title (incl cash) to the tooltip
                        tooltip:AddLine( "|n" );
                        tooltip:AddLine( UBI_ITEM_RECIPE_SOLD_BY:format( UberInventory_GetMoneyString( buyPrice ) ), 1.0, 1.0, 1.0 );

                        -- Add vendors to the tooltip
                        count = 0;
                        for idx, vendor in pairs( recipeData ) do
                            if ( count >= UBI_MAX_RECIPE_INFO ) then break; end;
                            name, area = strsplit( "|", UBI_Creatures[vendor] );
                            tooltip:AddDoubleLine( "  "..name, UBI_Zones[tonumber( area )] );
                            SetSmallerFont();
                            count = count + 1;
                        end;
                    end;

                    -- Add quest reward information for recipes
                    local recipeQuest = UBI_RecipeQuestRewards[itemId_num];
                    if ( UBI_Options["show_recipe_reward"] and recipeQuest ) then
                        -- Add section title to the tooltip
                        tooltip:AddLine( "|n" );
                        tooltip:AddLine( UBI_ITEM_RECIPE_REWARD_FROM, 1.0, 1.0, 1.0 );

                        -- Add quests to the tooltip
                        count = 0;
                        for idx, quest in pairs( recipeQuest ) do
                            if ( count >= UBI_MAX_RECIPE_INFO ) then break; end;
                            name, area = strsplit( "|", UBI_Quests[quest] )
                            tooltip:AddDoubleLine( "  "..name, UBI_Zones[tonumber( area )] );
                            SetSmallerFont();
                            count = count + 1;
                        end;
                    end;

                    -- Add drop information for recipes
                    local recipeDrop = UBI_RecipeDrops[itemId_num];
                    if ( UBI_Options["show_recipe_drop"] and recipeDrop ) then
                        -- Add section title to the tooltip
                        tooltip:AddLine( "|n" );
                        tooltip:AddLine( UBI_ITEM_RECIPE_DROP_BY, 1.0, 1.0, 1.0 );

                        -- Add quests to the tooltip
                        count = 0;
                        for idx, drop in pairs( recipeDrop ) do
                            if ( count >= UBI_MAX_RECIPE_INFO ) then break; end;
                            mob, rate = strsplit( ":", drop );
                            mob = tonumber( mob );
                            rate = tonumber( rate );
                            name, area = strsplit( "|", UBI_Creatures[mob] );
                            tooltip:AddDoubleLine( "  "..name.." ("..UBI_Droprates[rate]..")", UBI_Zones[tonumber( area )] );
                            SetSmallerFont();
                            count = count + 1;
                        end;
                    end;
                end;

                -- Add counts (if UBI_TooltipItem is populated then we are showing a tooltip from the main inventory frame)
                if ( UBI_TooltipItem ) then
                    -- Only add the info when we are not show data from the guildbank
                    if ( UBI_TooltipItem["bag_count"] and ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "current" or
                                                            UBI_LocationList[UBI_FILTER_LOCATIONS].type == "character" ) ) then
                        -- Add full item count information
                        AddLocationInfo( tooltip,
                                         UBI_TooltipItem["bag_count"],
                                         UBI_TooltipItem["bank_count"],
                                         UBI_TooltipItem["keyring_count"],
                                         UBI_TooltipItem["mailbox_count"],
                                         UBI_TooltipItem["equip_count"] );
                        totalCount = totalCount + UBI_TooltipItem["bag_count"] + UBI_TooltipItem["bank_count"] + UBI_TooltipItem["keyring_count"] + UBI_TooltipItem["mailbox_count"] + UBI_TooltipItem["equip_count"];
                        spacerAdded = true;
                    end;
                elseif ( UBI_Items[itemId] ) then
                    AddLocationInfo( tooltip,
                                     UBI_Items[itemId]["bag_count"],
                                     UBI_Items[itemId]["bank_count"],
                                     UBI_Items[itemId]["keyring_count"],
                                     UBI_Items[itemId]["mailbox_count"],
                                     UBI_Items[itemId]["equip_count"] );
                    totalCount = totalCount + UBI_Items[itemId]["bag_count"] + UBI_Items[itemId]["bank_count"] + UBI_Items[itemId]["keyring_count"] + UBI_Items[itemId]["mailbox_count"] + UBI_Items[itemId]["equip_count"];
                    spacerAdded = true;
                end;

                -- Add info from current character if viewing data from an alt
                if ( UBI_Items[itemId] and UBI_TooltipLocation and UBI_LocationList[UBI_FILTER_LOCATIONS].type ~= "current" ) then
                    -- Insert spacer if not already added
                    if ( not spacerAdded ) then
                        tooltip:AddLine( "|n" );
                        spacerAdded = true;
                    end;

                    -- Add the count and icon for the current alt
                    totalCount = totalCount + UBI_Items[itemId].total;
                    tooltip:AddDoubleLine( UBI_PLAYER, UBI_Items[itemId].total );
                    tooltip:AddTexture( UBI_LOCATION_TEXTURE[6] );
                end;

                -- Add information for alt characters
                local UBI_Characters = UBI_Characters;
                local record;
                for key, value in pairs( UBI_Characters ) do
                    if ( UBI_Data[UBI_REALM][value]["Items"][itemId] ) then
                        if ( value ~= strtrim( UBI_LocationList[UBI_TooltipLocation or 0].name ) ) then
                            -- Insert spacer if not already added
                            if ( not spacerAdded ) then
                                tooltip:AddLine( "|n" );
                                spacerAdded = true;
                            end;

                            -- Fix total
                            record = UBI_Data[UBI_REALM][value]["Items"][itemId];
                            record.total = record.bag_count + record.bank_count + record.keyring_count +
                                           record.mailbox_count + record.equip_count;
                            totalCount = totalCount + record.total;

                            -- Add the count and icon for the current alt
                            tooltip:AddDoubleLine( value, record.total );
                            tooltip:AddTexture( UBI_LOCATION_TEXTURE[6] );
                        end;
                    end;
                end;

                -- Add information for guildbanks
                local UBI_Guildbanks = UBI_Guildbanks;
                for key, value in pairs( UBI_Guildbanks ) do
                    if ( UBI_Guildbank[value]["Items"][itemId] ) then
                        -- Insert spacer if not already added
                        if ( not spacerAdded ) then
                            tooltip:AddLine( "|n" );
                            spacerAdded = true;
                        end;

                        -- Add the count and icon for the current guildbank
                        totalCount = totalCount + UBI_Guildbank[value]["Items"][itemId].total;
                        tooltip:AddDoubleLine( C_GREY..value..C_CLOSE, C_GREY..UBI_Guildbank[value]["Items"][itemId].total..C_CLOSE );
                        tooltip:AddTexture( UBI_LOCATION_TEXTURE[6] );
                    end;
                end;

                -- Add total itemcount
                if ( totalCount > 0 ) then
                    tooltip:AddDoubleLine( " ", C_YELLOW..UBI_SIGMA_ICON.." "..totalCount..C_CLOSE );
                end;
            end;
        end;

        tooltip.maxlines = tooltip:NumLines();
    end;

-- HOOK - Set new OnTooltipSetItem and OnHide script for Tooltip objects
    function UberInventory_HookTooltip( tooltip )
        -- From global to local
        local UBI_Hooks = UBI_Hooks;

        -- Store default script
        local tooltipName = tooltip:GetName();
        UBI_Hooks["OnTooltipSetItem"][tooltipName] = tooltip:GetScript( "OnTooltipSetItem" );
        UBI_Hooks["OnTooltipCleared"][tooltipName] = tooltip:GetScript( "OnTooltipCleared" );

        -- Set new script to handle OntooltipSetItem
        tooltip:SetScript( "OnTooltipSetItem", function( self, ... )
            -- From global to local
            local UBI_Hooks = UBI_Hooks;

            -- Get tooltip name
            local tooltipName = self:GetName();

            -- Call default script
            if ( UBI_Hooks["OnTooltipSetItem"][tooltipName] ) then
                UBI_Hooks["OnTooltipSetItem"][tooltipName]( self, ... );
            end;

            -- Call new script (adds the item information)
            UberInventory_AddItemInfo( self );

            -- Turn on UberInventory indicator
            self.UBI_InfoAdded = true;
        end );

        -- Set new script to handle OnTooltipCleared
        tooltip:SetScript( "OnTooltipCleared", function( self, ... )
            -- From global to local
            local UBI_Hooks = UBI_Hooks;

            -- Get tooltip name
            local tooltipName = self:GetName();

            -- Force reset of fonts (maxlines is a custom attribute added within the UberInventory_AddItemInfo function)
            if ( self.maxlines ) then
                local txtLeft, txtRight;
                for i = 1, self.maxlines do
                    txtLeft = _G[self:GetName().."TextLeft"..i];
                    txtRight = _G[self:GetName().."TextRight"..i];
                    if ( txtLeft ) then txtLeft:SetFontObject( GameTooltipText ); end;
                    if ( txtRight ) then txtRight:SetFontObject( GameTooltipText ); end;
                end;
            end;

            -- Call default script
            if ( UBI_Hooks["OnTooltipCleared"][tooltipName] ) then
                UBI_Hooks["OnTooltipCleared"][tooltipName]( self, ... );
            end;

            -- Turn off UberInventory indicator
            self.UBI_InfoAdded = false;
        end );
     end;

-- Get item information and store it
    function UberInventory_Item( bagID, slotID, location )
        -- From global to local
        local UBI_Items = UBI_Items;

        -- Initialize
        local bankCount = 0;
        local bagCount = 0;
        local keyCount = 0;
        local mailCount = 0;
        local equipCount = 0;
        local itemLink = nil;

        -- Get itemLink
        if ( location == "mailbox" ) then
            itemLink = GetInboxItemLink( bagID, slotID );
        elseif ( location == "equip" ) then
            itemLink = GetInventoryItemLink( "player", slotID );
        else
            itemLink = GetContainerItemLink( bagID, slotID );
        end;

        -- Retrieve item information
        if ( itemLink ) then
            -- Initialize
            local itemCount = 0;

            -- Fetch item information (itemLink skipped from GetItemInfo, does not always work at this stage (Quest related items))
            local itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );
            local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit( ":", itemLink );

            if ( location == "mailbox" ) then
                _, _, itemCount = GetInboxItem( bagID, slotID );
            elseif ( location == "equip" ) then
                itemCount = 1;
            else
                _, itemCount = GetContainerItemInfo( bagID, slotID );
            end;

            -- Get existing item counts or activate sorting (new item is added)
            if ( UBI_Items[itemId] ) then
                bankCount = UBI_Items[itemId]["bank_count"];
                bagCount = UBI_Items[itemId]["bag_count"];
                keyCount = UBI_Items[itemId]["keyring_count"];
                mailCount = UBI_Items[itemId]["mailbox_count"];
                equipCount = UBI_Items[itemId]["equip_count"];
            end;

            -- Sort out itemCount
            if ( location == "bank" ) then
                bankCount = bankCount + itemCount;
            elseif ( location == "bag" ) then
                bagCount = bagCount + itemCount;
            elseif ( location == "keyring" ) then
                keyCount = keyCount + itemCount;
            elseif ( location == "mailbox" ) then
                mailCount = mailCount + itemCount;
            elseif ( location == "equip" ) then
                equipCount = equipCount + itemCount;
            end;
            local totalCount = bagCount + bankCount + keyCount + mailCount + equipCount;

            -- Store the item information
            UBI_Items[itemId] = { ["itemid"] = tonumber( itemId ),
                                  ["name"] = itemName,
                                  ["quality"] = itemQuality,
                                  ["level"] = itemLevel,
                                  ["bag_count"] = bagCount,
                                  ["bank_count"] = bankCount,
                                  ["keyring_count"] = keyCount,
                                  ["mailbox_count"] = mailCount,
                                  ["equip_count"] = equipCount,
                                  ["type"] = itemType,
                                  ["subtype"] = itemSubType,
                                  ["total"] = totalCount };
        end;
    end;

-- Handle Bag items
    function UberInventory_Save_Bag()
        -- Initialize
        local slotCount, freeCount = 0, 0;

        -- Reset bag counts
        UberInventory_ResetCount( "bag" );

        -- Traverse bags
        for bagID = 0, 4 do
            if ( GetContainerNumSlots( bagID ) ) then
                slotCount = slotCount + GetContainerNumSlots( bagID );
                freeCount = freeCount + GetContainerNumFreeSlots( bagID );
                for slotID = 1, GetContainerNumSlots( bagID ) do
                    UberInventory_Item( bagID, slotID, "bag" );
                end;
            end;
        end;

        -- Save slot count
        UBI_Options["bag_max"] = slotCount;
        UBI_Options["bag_free"] = freeCount;
    end;

-- Handle Bank items
    function UberInventory_Save_Bank()
        -- Intialize
        local slotCount, freeCount = 0, 0;

        if ( UBI_BANK_OPEN ) then
            -- Reset bank counts
            UberInventory_ResetCount( "bank" );

            -- Traverse bank slots
            slotCount = GetContainerNumSlots( BANK_CONTAINER );
            for slotID = 1, GetContainerNumSlots( BANK_CONTAINER ) do
                if ( not GetContainerItemLink( BANK_CONTAINER, slotID ) ) then
                    freeCount = freeCount + 1;
                end;
                UberInventory_Item( BANK_CONTAINER, slotID, "bank" );
            end;

            -- Traverse bank bags
            for bagID = 5, 11 do
                if ( GetContainerNumSlots( bagID ) ) then
                    slotCount = slotCount + GetContainerNumSlots( bagID );
                    freeCount = freeCount + GetContainerNumFreeSlots( bagID );
                    for slotID = 1, GetContainerNumSlots( bagID ) do
                        UberInventory_Item( bagID, slotID, "bank" );
                    end;
                end;
            end;

            -- Save slot count
            UBI_Options["bank_max"] = slotCount;
            UBI_Options["bank_free"] = freeCount;
        end;
    end;

-- Handle Keyring items
    function UberInventory_Save_Keyring()
        -- Reset keyring counts
        UberInventory_ResetCount( "keyring" );

        -- Traverse keyring
        if ( HasKey() == 1 ) then
            for slotID = 1, GetContainerNumSlots( KEYRING_CONTAINER ) do
                UberInventory_Item( KEYRING_CONTAINER, slotID, "keyring" );
            end;
        end;
    end;

-- Handle Mailbox items
    function UberInventory_Save_Mailbox()
        -- Initialize
        local today = time();
        local UBI_Options = UBI_Options;
        local UBI_Money = UBI_Money;
        local mailSender, mailSubject, cashAttached, daysLeft, itemAttached;
        local mailText, mailTexture, mailTakeable, mailInvoice;
        local invoiceType, bid, buyout, deposit, consignment;

        if ( UBI_MAILBOX_OPEN and not UBI_COLLECTING_CASH ) then
            -- Reset mailbox cash
            UBI_Money["mail"] = 0;

            -- Reset expiration date
            if ( UBI_Options["mail_expires"] ) then wipe( UBI_Options["mail_expires"] ) else UBI_Options["mail_expires"] = {}; end;

            -- Reset mailbox counts
            UberInventory_ResetCount( "mailbox" );

            -- Warn if the max of 50 is stored in the mailbox
            if ( UBI_Options["warn_fullmailbox"] ) then
                if ( GetInboxNumItems() == UBI_MAIL_MAX and not UBI_MAIL_WARNING ) then
                    PlaySound("igQuestLogAbandonQuest");
                    UIErrorsFrame:AddMessage( UBI_MAIL_MAX1.."\n"..UBI_MAIL_MAX2, 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME );
                    UberInventory_Message( UBI_MAIL_MAX1 );
                    UberInventory_Message( UBI_MAIL_MAX2 );
                    UBI_MAIL_WARNING = true;
                end;
            end;

            -- Traverse mailbox (Reverse order)
            for mailID = GetInboxNumItems(), 1, -1 do
                -- Mail info
                _, _, mailSender, mailSubject, cashAttached, _, daysLeft, itemAttached = GetInboxHeaderInfo( mailID );
                daysLeft = today + ( floor( daysLeft ) * UBI_SEC_IN_DAY );

                -- Register all mails stored in mailbox stored by expiration date (exclude sales pending mails)
        if ( itemAttached or cashAttached > 0 ) then
          UBI_Options["mail_expires"][daysLeft] = ( UBI_Options["mail_expires"][daysLeft] or 0 ) + 1;
        end;

                -- Handle attachment
                if ( itemAttached ) then
                    for attachID = 1, ATTACHMENTS_MAX do
                        UberInventory_Item( mailID, attachID, "mailbox" );
                    end;
                end;

                -- Handle money
                if ( cashAttached > 0 ) then
                    if ( UBI_Options["take_money"] ) then
                        -- Indicate we are collecting cash
                        UBI_COLLECTING_CASH = true;

                        -- Add task to take inbox money (if not items are attached this also deletes the mail)
                        TaskHandlerLib:AddTask( "UberInventory", UBI_Task_CollectCash, mailID );
                    else
                        -- Record the mailbox cash
                        UBI_Money["mail"] = UBI_Money["mail"] + cashAttached;
                    end;
                end;

                -- Update wallet/mail info (if mainframe is shown)
                UberInventory_WalletMailCashInfo();

                -- At the last (actually the first) mail add task to turn off cash collecting
                TaskHandlerLib:AddTask( "UberInventory", UBI_Task_DoneCollectingCash, UBI_COLLECTING_CASH );
            end;
        end;
    end;

-- Handle Equiped items
    function UberInventory_Save_Equipped()
        -- Initialize
        local slotID;

        -- Reset Equiped counts
        UberInventory_ResetCount( "equip" );

        -- Traverse equip slots
        for key, value in pairs( UBI_EQUIP_SLOTS ) do
            slotID, _ = GetInventorySlotInfo( value );
            UberInventory_Item( nil, slotID, "equip" );
        end;
    end;

-- Handle Guildbank items
    function UberInventory_Save_Guildbank( event )
    	-- Exit if already saving gb data
		if ( UBI_PROCESSING_GB ) then return; end;

        -- From global to local
        local UBI_Guildbank = UBI_Guildbank;

        -- Initialize
        local itemCount = 0;
        local name, icon, isViewable, canDeposit, numWithdrawals, remainingWithdrawals, count, itemLink;
        local itemName, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture;
        local itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId;

        if ( UBI_GUILDBANK_OPEN ) then
			-- indicate we are already saving gb data
			UBI_PROCESSING_GB = true;

            -- Start out to assume we have view access to all guildbank tabs
            UBI_GUILDBANK_VIEWACCESS = true;

            -- Clear the existing data
            wipe( UBI_Guildbank[UBI_GUILD]["Items"] );

            -- Traverse all guildbank tabs and slots
            for gbTab = 1, GetNumGuildBankTabs() do
                -- Retrieve data from the server
                if ( event ~= "GUILDBANKBAGSLOTS_CHANGED" ) then
                    QueryGuildBankTab( gbTab );
                end;

                -- Can we view the content of all guildbank tabs?
                if ( not IsGuildLeader( UBI_PLAYER ) ) then
                    name, icon, isViewable, canDeposit, numWithdrawals, remainingWithdrawals = GetGuildBankTabInfo( gbTab );
                    if ( not isViewable ) then
                        UBI_GUILDBANK_VIEWACCESS = false;
                    end;
                end;

                for gbSlot = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
                    -- Reset
                    itemCount = 0;

                    -- Get slot information
                    _, count, _ = GetGuildBankItemInfo( gbTab, gbSlot );
                    itemLink = GetGuildBankItemLink( gbTab, gbSlot );

                    if ( itemLink ) then
                        -- Fetch item information
                        itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );
                        _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit( ":", itemLink );

                        -- Get existing item count
                        if ( UBI_Guildbank[UBI_GUILD]["Items"][itemId] ) then
                            itemCount = UBI_Guildbank[UBI_GUILD]["Items"][itemId]["count"];
                        end;

                        -- Store the item information
                        UBI_Guildbank[UBI_GUILD]["Items"][itemId] = { ["itemid"] = tonumber( itemId ),
                                                                      ["name"] = itemName,
                                                                      ["quality"] = itemQuality,
                                                                      ["level"] = itemLevel,
                                                                      ["count"] = itemCount + count,
                                                                      ["type"] = itemType,
                                                                      ["subtype"] = itemSubType,
                                                                      ["total"] = itemCount + count };
                    end;
                end;
            end;

            -- Done saving gb data
            UBI_PROCESSING_GB = false;
        end;

        -- Rebuild list of guildbanks
        UberInventory_GetGuildbanks();

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;
    end;

-- Handle Guildbank items
    function UberInventory_Receive_Guildbank()
        -- From global to local
        local UBI_Guildbank = UBI_Guildbank;
        local UBI_GBData = UBI_GBData;

        -- Init guildbank structures (if needed)
        UberInventory_Guildbank_Init()

        -- Show receiving data
        UberInventory_Message( UBI_NEW_GBDATA:format( UBI_GBSender ) );

        -- Clear the existing data
        wipe( UBI_Guildbank[UBI_GUILD]["Items"] );

        -- Traverse all guildbank tabs and slots
        local itemLink, itemName, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture;
        for key, value in pairs( UBI_GBData ) do
            -- Fetch item information
            itemLink = UberInventory_GetLink( value.itemid );
            itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );

            -- Store the item information
            UBI_Guildbank[UBI_GUILD]["Items"][value.itemid] = { ["itemid"] = tonumber( value.itemid ),
                                                                ["name"] = value.name or itemName or UBI_ITEM_UNCACHED,
                                                                ["quality"] = value.quality or itemQuality,
                                                                ["level"] = value.level or itemLevel or 0,
                                                                ["count"] = value.count,
                                                                ["type"] = value.type or itemType or UBI_ITEM_UNCACHED,
                                                                ["subtype"] = value.subtype or itemSubType or UBI_ITEM_UNCACHED,
                                                                ["total"] = value.count };
        end;

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;

        -- Track Guildbank last visit date/time
        UBI_Options["guildbank_visit"] = time();

        -- Clear sender
        UBI_GBSender = nil;
    end;

-- Send Guildbank data to other guild members
    function UberInventory_Send_Guildbank()
        -- From global to local
        local UBI_Guildbank = UBI_Guildbank;
        local sendData;

        -- Get number of online guild members
        local numOnline, numTotal = GetNumGuildMembers();

        -- Only send data when we have view access to all guildbank tabs
        if ( UBI_GUILDBANK_VIEWACCESS and numOnline > 1 ) then
            -- Send start message
            TaskHandlerLib:AddTask( "UberInventory", UBI_Task_SendStart, UBI_VERSION );

            -- Send the actual guildbank data
            for key, record in pairs( UBI_Guildbank[UBI_GUILD]["Items"] ) do
                -- Create data string
                sendData = key.." "..
                           record["count"].." "..
                           strgsub( record["name"], " ", "_" ).." "..
                           strgsub( record["type"], " ", "_" ).." "..
                           strgsub( record["subtype"], " ", "_" ).." "..
                           record["level"].." "..
                           record["quality"];

                -- Send data string
                TaskHandlerLib:AddTask( "UberInventory", UBI_Task_SendData, sendData );
            end;

            -- Send cash
            TaskHandlerLib:AddTask( "UberInventory", UBI_Task_SendCash, UBI_Guildbank[UBI_GUILD]["Cash"] );

            -- Send done message
            TaskHandlerLib:AddTask( "UberInventory", UBI_Task_SendEnd );
        end;
    end;

-- Check for expiring mails (only when the option warn_mailexpire is on)
    function UberInventory_MailExpirationCheck()
        -- From global to local
        local UBI_Data = UBI_Data;
        local realm, daysLeft;

        if ( UBI_Options["warn_mailexpire"] ) then
            -- Check mail expiration for all characters from the current realm
            for key, record in pairs( UBI_Data ) do
                realm = key;

                for player, value in pairs( UBI_Data[realm] ) do
                    if ( player ~= "Guildbank" and value["Options"]["mail_expires"] ) then
                        for key, record in pairs( value["Options"]["mail_expires"] ) do
                            -- Get number of days till expiration (negative number are good)
                            daysLeft = UberInventory_DaysSince( key );

                            -- If daysleft less then zero, mail expires into the future, positive numbers means the mail are already expired
                            if ( daysLeft < 0 ) then
                                daysLeft = abs( daysLeft );
                                if ( daysLeft <= UBI_MAIL_EXPIRE_WARNING ) then
                                    UberInventory_Message( UBI_MAIL_EXPIRES:format( player, realm, record, daysLeft ) );
                                end;
                            else
                                UberInventory_Message( UBI_MAIL_LOST:format( player, realm, record, daysLeft ) );
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;

-- Cleanup inventory
    function UberInventory_Cleanup()
        -- From global to local
        local UBI_Items = UBI_Items;

        -- Remove items for which bag_count, bank_count, keyring_count, mailbox_count and equip_count equals 0
        for key, record in pairs( UBI_Items ) do
            if ( record.name ) then
                if ( record.bag_count + record.bank_count + record.keyring_count + record.mailbox_count + record.equip_count == 0 ) then
                    -- Remove the item
                    UBI_Items[key] = nil;
                end;
            else
                -- Remove the item if name is not found
                UBI_Items[key] = nil;
            end;
        end;
    end;

-- Update inventory
    function UberInventory_UpdateInventory( location )
        -- Update the inventory
        if ( location == "bank" or location == "all" ) then
            UberInventory_Save_Bank();
        end;
        if ( location == "bag" or location == "all" ) then
            UberInventory_Save_Bag();
            UberInventory_Save_Keyring();
        end;
        if ( location == "mail" or location == "all" ) then
            UberInventory_Save_Mailbox();
        end;
        if ( location == "all" ) then
            UberInventory_Save_Equipped();
        end;

        -- Store last update info
        UBI_SCAN_DATE = time();

        -- Cleanup the inventory (remove zero-count items)
        UberInventory_Cleanup();

        -- Get cash information
        UberInventory_GetMoney();

        -- Update wallet/mail info (if mainframe is shown)
        UberInventory_WalletMailCashInfo();

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;
    end;

-- Handle UBI initialization
    function UberInventory_Init()
        -- Initiliaze all parts of the saved variable
        if (not UBI_Data) then
            UBI_Data = {};
        end;

        if (not UBI_Data[UBI_REALM]) then
            UBI_Data[UBI_REALM] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]) then
            UBI_Data[UBI_REALM][UBI_PLAYER] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Options"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Options"] = { ["version"] = UBI_VERSION,
                                                           ["show_money"] = true,
                                                           ["show_balance"] = true,
                                                           ["show_sell_prices"] = true,
                                                           ["show_recipe_prices"] = true,
                                                           ["show_recipe_reward"] = true,
                                                           ["show_recipe_drop"] = true,
                                                           ["show_minimap"] = true,
                                                           ["show_highlight"] = true,
                                                           ["send_gb_data"] = true,
                                                           ["receive_gb_data"] = true,
                                                           ["warn_mailexpire"] = true,
                                                           ["warn_fullmailbox"] = true,
                                                           ["bag_max"] = 0,
                                                           ["bag_free"] = 0,
                                                           ["bank_max"] = 0,
                                                           ["bank_free"] = 0,
                                                           ["minimap"] = 315,
                                                           ["alpha"] = 1,
                                                           ["take_money"] = false,
                                                           ["faction"] = UBI_FACTION,
                                                           ["mail_expires"] = {},
                                                         };
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Items"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Items"] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Money"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Money"] = { ["mail"] = 0,
                                                         ["Cash"] = 0 };
        end;

        if (not UBI_Data[UBI_REALM]["Guildbank"]) then
            UBI_Data[UBI_REALM]["Guildbank"] = {};
        end;

        -- Create global variables for easy reference
        UBI_Money = UBI_Data[UBI_REALM][UBI_PLAYER]["Money"];
        UBI_Items = UBI_Data[UBI_REALM][UBI_PLAYER]["Items"];
        UBI_Options = UBI_Data[UBI_REALM][UBI_PLAYER]["Options"];
        UBI_Guildbank = UBI_Data[UBI_REALM]["Guildbank"];

        -- Upgrade data structures
        UberInventory_Upgrade();

        -- Build list of alts and guildbanks
        UberInventory_GetAlts();
        UberInventory_GetGuildbanks();

        -- Build the inventory
        UberInventory_UpdateInventory( "all" );

        -- Initialize Options dialog
        UberInventoryOptionsFrame:Show();
    end;

-- Initialize guildbank data structure
    function UberInventory_Guildbank_Init()
        if ( IsInGuild() ) then
            if ( UBI_GUILD and not UBI_Guildbank[UBI_GUILD] ) then
                UBI_Guildbank[UBI_GUILD] = {};
                UBI_Guildbank[UBI_GUILD]["Items"] = {};
                UBI_Guildbank[UBI_GUILD]["Cash"] = 0;
                UBI_Guildbank[UBI_GUILD]["Faction"] = UBI_FACTION;
            end;
        end;
    end;

-- Register events
    function UberInventory_RegisterEvents()
        for key, value in pairs( UBI_TRACKED_EVENTS ) do
            this:RegisterEvent( value );
        end;
    end;

-- Handle startup of the addon
    function UberInventory_OnLoad()
        -- Install hooks
        UberInventory_Install_Hooks();

        -- Show startup message
        UberInventory_Message( UBI_STARTUP_MESSAGE, false );

        -- Register slash commands
        SlashCmdList["UBI"] = UberInventory_SlashHandler;
        SLASH_UBI1 = "/ubi";
        SLASH_UBI2 = "/uberinventory";

        -- Register events to be monitored
        UberInventory_RegisterEvents();

        -- Correct quality list
        tinsert( UBI_QUALITY, 1, UBI_ALL_QUALITIES );
    end;

-- Install all of the hooks used by UberInventory
    function UberInventory_Install_Hooks()
        -- Hook the Tooltips (OnTooltipSetItem, OnTooltipCleared)
        UberInventory_HookTooltip( GameTooltip );
        UberInventory_HookTooltip( ItemRefTooltip );
        UberInventory_HookTooltip( ShoppingTooltip1 );
        UberInventory_HookTooltip( ShoppingTooltip2 );
        UberInventory_HookTooltip( ShoppingTooltip3 );
    end;

-- Handle events sent to the addon
    function UberInventory_OnEvent( self, event, ... )

        -- Load data from previous session
        if ( event == "VARIABLES_LOADED" ) then
            -- Execute main initialization code
            UberInventory_Init();

            -- Check bank visit
            if ( UberInventory_DaysSince( UBI_Options["bank_visit"] ) >= UBI_BANK_VISIT_INTERVAL ) then
                UberInventory_Message( UBI_LAST_BANK_VISIT:format( UBI_Options["bank_visit"] ) );
            else
                if ( not UBI_Options["bank_visit"] ) then UberInventory_Message( UBI_VISIT_BANK ); end;
            end;

            -- Check guildbank visit
            if ( IsInGuild() ) then
                if ( UberInventory_DaysSince( UBI_Options["guildbank_visit"] ) >= UBI_GUILDBANK_VISIT_INTERVAL ) then
                    UberInventory_Message( UBI_LAST_GUILDBANK_VISIT:format( UBI_Options["guildbank_visit"] ) );
                else
                    if ( not UBI_Options["guildbank_visit"] ) then UberInventory_Message( UBI_VISIT_GUILDBANK ); end;
                end;
            end;

            -- Check mailbox visit
            if ( UberInventory_DaysSince( UBI_Options["mail_visit"] ) >= UBI_MAILBOX_VISIT_INTERVAL ) then
                UberInventory_Message( UBI_LAST_MAILBOX_VISIT:format( UBI_Options["mail_visit"] ) );
            else
                if ( not UBI_Options["mail_visit"] ) then UberInventory_Message( UBI_VISIT_MAILBOX ); end;
            end;

            -- Check for mail that are about to expire
            UberInventory_MailExpirationCheck();

            -- Show or hide the minimap icon
            UberInventory_Minimap_Toggle();

            -- Assign proper location for minimap
            UberInventory_Minimap_Update( UBI_Options["minimap"] );

            -- Set transparency
            UberInventory_Change_Alpha( UBI_Options["alpha"] );

            -- Broadcast current version to the guild
            UberInventory_SendVersion();
            return;
        end;

        -- Update info inventory change (enters the world, enters/leaves an instance, or respawns at a graveyard, reloadui)
        if ( event == "PLAYER_ENTERING_WORLD" ) then
            -- Update balance frame
            UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

            -- Get guild name
            UBI_GUILD = GetGuildInfo( "player" );
            UberInventory_Guildbank_Init();

            UberInventory_UpdateInventory( "all" );
            return;
        end;

        if ( event == "UNIT_INVENTORY_CHANGED" and arg1 == "player" ) then
            UberInventory_UpdateInventory( "all" );
            return;
        end;

        -- After a money event show the current cash amount
        if ( event == "PLAYER_MONEY" ) then
            -- Save current cash balance
            UberInventory_GetMoney();

            -- Update wallet/mail info (if mainframe is shown)
            UberInventory_WalletMailCashInfo();

            -- Show message
            if ( UBI_Options["show_money"] ) then
                UberInventory_Message( UBI_MONEY_MESSAGE:format( UberInventory_GetMoneyString( GetMoney() ) ) );
            end;
            return;
        end;

        -- Bank closed
        if ( event == "BANKFRAME_CLOSED" ) then
            UBI_BANK_OPEN = false;
            return;
        end;

        -- Bank opened
        if ( event == "BANKFRAME_OPENED" ) then
            UBI_BANK_OPEN = true;
            UBI_Options["bank_visit"] = time();
            UberInventory_UpdateInventory( "bank" );
            return;
        end;

        -- Items in the bank have changed
        if ( event == "PLAYERBANKSLOTS_CHANGED" or event == "PLAYERBANKBAGSLOTS_CHANGED" ) then
            UberInventory_UpdateInventory( "bank" );
            return;
        end;

        -- Guildbank closed (send guildbank data using the chat system)
        if ( event == "GUILDBANKFRAME_CLOSED" and UBI_GUILDBANK_OPEN ) then
            -- Store guildbank cash
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Start sending guildbank data
            if ( UBI_Options["send_gb_data"] ) then
                UberInventory_Send_Guildbank();
            end;

            UBI_GUILDBANK_OPEN = false;
            return;
        end;

        -- Guild cash changed (When does this event occur?)
        if ( event == "GUILDBANK_UPDATE_MONEY" and IsInGuild() ) then
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Update cash info on screen
            if ( UberInventoryFrame:IsVisible() ) then
                UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
            end;
            return;
        end;

        -- Guildbank opened
        if ( event == "GUILDBANKFRAME_OPENED" ) then
            -- Track Guildbank open
            UBI_GUILDBANK_OPEN = true;

            -- Store guildbank cash
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Track Guildbank last visit date/time
            UBI_Options["guildbank_visit"] = time();

            -- Store Guildbank data
            UberInventory_Save_Guildbank( event );
            return;
        end;

        -- Update Guildbank information
        if ( ( event == "GUILDBANK_UPDATE_TABS" or event == "GUILDBANKBAGSLOTS_CHANGED" ) and UBI_GUILDBANK_OPEN ) then
            -- Send start message
            TaskHandlerLib:AddTask( "UberInventory", UberInventory_Save_Guildbank, event );
--            UberInventory_Save_Guildbank( event );
            return;
        end;

        -- Mailbox opened
        if ( event == "MAIL_SHOW" ) then
            UBI_MAILBOX_OPEN = true;
            UBI_Options["mail_visit"] = time();
            UberInventory_UpdateInventory( "mail" );
            return;
        end;

        -- Mailbox changes
        if ( event == "MAIL_INBOX_UPDATE" or event == "UPDATE_PENDING_MAIL" ) then
            UberInventory_UpdateInventory( "mail" );
            return;
        end;

        -- Mailbox closed
        if ( event == "MAIL_CLOSED" ) then
            UBI_MAILBOX_OPEN = false;
            UBI_MAIL_WARNING = false;
            return;
        end;

        -- Items in the bags have changed
        if ( event == "BAG_UPDATE" and arg1 >= -2 ) then
            -- Update inventory
            UberInventory_UpdateInventory( "bag" );
            return;
        end;

        -- Handle UberInventory related guildchat messages
        if ( event == "CHAT_MSG_ADDON" and arg1 == "UBI" and arg4 ~= UBI_PLAYER and UBI_Options["receive_gb_data"] ) then
            if ( arg2:find( "UBI:VERSION" ) ) then
                -- Extract version number from the data string
                local _, gbVersion = strsplit( " ", arg2 );

                -- Check if there is a more recent version of UberInventory
                if ( tonumber( gbVersion ) > ( tonumber( UBI_VERSION ) or 0 ) and not UBI_VERSION_WARNING ) then
                    UberInventory_Message( UBI_NEW_VERSION:format( arg4, gbVersion ) );
                    UBI_VERSION_WARNING = true;
                end;

                -- If this version is more recent send message to other users using this addon
                if ( tonumber( gbVersion ) < ( tonumber( UBI_VERSION ) or 0 ) ) then
                    UberInventory_SendVersion();
                end;
            elseif ( arg2:find( "UBI:GBSTART" ) and not UBI_GUILDBANK_OPEN ) then
                -- Start receiving Guildbank data, reset on each new receive from player.
                -- This way only the most recent data should be received.
                -- Only data from equal versions are allowed
                wipe( UBI_GBData );
                UBI_GBSender = arg4;
            elseif ( arg2:find( "UBI:GBITEM" ) and not UBI_GUILDBANK_OPEN ) then
                -- Receiving Guildbank data (item), only receive data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    local _, itemID, itemCount, itemName, itemType, itemSubtype, itemLevel, itemQuality = strsplit( " ", arg2 );
                    UBI_GBData[itemID] = { ["itemid"] = tonumber( itemID ),
                                           ["count"] = tonumber( itemCount ),
                                           ["name"] = strgsub( itemName, "_", " " ),
                                           ["type"] = strgsub( itemType, "_", " " ),
                                           ["subtype"] = strgsub( itemSubtype, "_", " " ),
                                           ["level"] = tonumber( itemLevel ),
                                           ["quality"] = tonumber( itemQuality ) };
                end;
            elseif ( arg2:find( "UBI:GBCASH" ) and not UBI_GUILDBANK_OPEN ) then
                -- Receiving Guildbank cash, only receive data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    -- Init guildbank structures (if needed)
                    UberInventory_Guildbank_Init();

                    local _, gbCash = strsplit( " ", arg2 );
                    UBI_Guildbank[UBI_GUILD]["Cash"] = tonumber( gbCash );

                    -- Update cash info on screen
                    if ( UberInventoryFrame:IsVisible() ) then
                        UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
                    end;
                end;
            elseif ( arg2:find( "UBI:GBEND" ) and not UBI_GUILDBANK_OPEN ) then
                -- Stop receiving Guildbank data and process the received data, only process data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    UberInventory_Receive_Guildbank();
                end;
            end;
        end;

        -- Rebuild list of guildbanks
        if ( UBI_GUILD ~= GetGuildInfo( "player" ) ) then
            UBI_GUILD = GetGuildInfo( "player" );
            UberInventory_Guildbank_Init();
            UberInventory_GetGuildbanks();
            UberInventory_Locations_Init();
        end;

    end;

-- Show or hide the balance frame
    function UberInventory_ShowBalanceFrame( showFrame )
        if ( showFrame ) then
            UberInventoryBalanceFrame:Show();
            MoneyFrame_Update( "UberInventoryBalanceFrameMoneyFrame", GetMoney() );
        else
            UberInventoryBalanceFrame:Hide();
        end;
    end;

-- Update minimap button
    function UberInventory_Minimap_Update( angle )
        UberInventoryMinimap:SetPoint( "TOPLEFT",
                                       "Minimap",
                                       "TOPLEFT",
                                       54 - (78 * cos( angle )),
                                       (78 * sin( angle )) - 55 );
    end;

-- Update minimap button
    function UberInventory_Minimap_Toggle()
        if ( UBI_Options["show_minimap"]) then
            UberInventoryMinimap:Show();
        else
            UberInventoryMinimap:Hide();
        end;
    end;

-- Update transparency of frames
    function UberInventory_Change_Alpha( alpha )
        UberInventoryFrameTexture:SetAlpha( alpha );
        UberInventoryFrameBorderTexture:SetAlpha( alpha );
    end;

-- Init frame options
    function UberInventory_SetOptions( frame )
        -- Initiliaze
        local objPrefix = "Setting";
        local UBI_Options = UBI_Options;

        -- Apply settings to the frame
        _G[ objPrefix.."Money" ]:SetChecked( UBI_Options["show_money"] );
        _G[ objPrefix.."Balance" ]:SetChecked( UBI_Options["show_balance"] );
        _G[ objPrefix.."SellPrice" ]:SetChecked( UBI_Options["show_sell_prices"] );
        _G[ objPrefix.."RecipePrice" ]:SetChecked( UBI_Options["show_recipe_prices"] );
        _G[ objPrefix.."QuestReward" ]:SetChecked( UBI_Options["show_recipe_reward"] );
        _G[ objPrefix.."RecipeDrop" ]:SetChecked( UBI_Options["show_recipe_drop"] );
        _G[ objPrefix.."Highlight" ]:SetChecked( UBI_Options["show_highlight"] );
        _G[ objPrefix.."InboxMoney" ]:SetChecked( UBI_Options["take_money"] );
        _G[ objPrefix.."ShowMinimap" ]:SetChecked( UBI_Options["show_minimap"] );
        _G[ objPrefix.."Minimap" ]:SetValue( UBI_Options["minimap"] );
        _G[ objPrefix.."Alpha" ]:SetValue( UBI_Options["alpha"] );
        _G[ objPrefix.."SendGuildbankData" ]:SetChecked( UBI_Options["send_gb_data"] );
        _G[ objPrefix.."ReceiveGuildbankData" ]:SetChecked( UBI_Options["receive_gb_data"] );
        _G[ objPrefix.."WarnMailExpire" ]:SetChecked( UBI_Options["warn_mailexpire"] );
        _G[ objPrefix.."WarnFullMailbox" ]:SetChecked( UBI_Options["warn_fullmailbox"] );
    end;

-- Save frame options
    function UberInventory_SaveOptions( frame )
        -- Initiliaze
        local objPrefix = "Setting";
        local UBI_Options = UBI_Options;

        -- Save settings from the frame
        UBI_Options["show_money"] = _G[ objPrefix.."Money" ]:GetChecked();
        UBI_Options["show_balance"] = _G[ objPrefix.."Balance" ]:GetChecked();
        UBI_Options["show_sell_prices"] = _G[ objPrefix.."SellPrice" ]:GetChecked();
        UBI_Options["show_recipe_prices"] = _G[ objPrefix.."RecipePrice" ]:GetChecked();
        UBI_Options["show_recipe_reward"] = _G[ objPrefix.."QuestReward" ]:GetChecked();
        UBI_Options["show_recipe_drop"] = _G[ objPrefix.."RecipeDrop" ]:GetChecked();
        UBI_Options["show_highlight"] = _G[ objPrefix.."Highlight" ]:GetChecked();
        UBI_Options["take_money"] = _G[ objPrefix.."InboxMoney" ]:GetChecked();
        UBI_Options["show_minimap"] = _G[ objPrefix.."ShowMinimap" ]:GetChecked();
        UBI_Options["minimap"] = _G[ objPrefix.."Minimap" ]:GetValue();
        UBI_Options["alpha"] = _G[ objPrefix.."Alpha" ]:GetValue();
        UBI_Options["send_gb_data"] = _G[ objPrefix.."SendGuildbankData" ]:GetChecked();
        UBI_Options["receive_gb_data"] = _G[ objPrefix.."ReceiveGuildbankData" ]:GetChecked();
        UBI_Options["warn_mailexpire"] = _G[ objPrefix.."WarnMailExpire" ]:GetChecked();
        UBI_Options["warn_fullmailbox"] = _G[ objPrefix.."WarnFullMailbox" ]:GetChecked();
    end;

-- Apply setting to the UI
    function UberInventory_UpdateUI()
        -- Initiliaze
        local UBI_Options = UBI_Options;

        -- Hide or show balance frame based on the current settings (look for a more elegant solution)
        UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

        -- Show or hide minimap icon
        UberInventory_Minimap_Toggle();

        -- Restore or set minimap location
        UberInventory_Minimap_Update( UBI_Options["minimap"] );

        -- Restore or set transparency
        UberInventory_Change_Alpha( UBI_Options["alpha"] );
    end;

-- Revert all options to default values
    function UberInventory_SetDefaults( frame )
        -- Initiliaze
        local UBI_Options = UBI_Options;

        -- Restore defaults
        UBI_Options["show_money"] = true;
        UBI_Options["show_balance"] = true;
        UBI_Options["show_sell_prices"] = true;
        UBI_Options["show_recipe_prices"] = true;
        UBI_Options["show_recipe_drop"] = true;
        UBI_Options["show_minimap"] = true;
        UBI_Options["minimap"] = 315;
        UBI_Options["alpha"] = 1;
        UBI_Options["take_money"] = false;
        UBI_Options["send_gb_data"] = true;
        UBI_Options["receive_gb_data"] = true;
        UBI_Options["warn_mailexpire"] = true;
        UBI_Options["warn_fullmailbox"] = true;

        -- Hide or show balance frame based on the current settings (look for a more elegant solution)
        UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

        -- Show or hide minimap icon
        UberInventory_Minimap_Toggle();

        -- Restore or set minimap location
        UberInventory_Minimap_Update( UBI_Options["minimap"] );

        -- Restore or set transparency
        UberInventory_Change_Alpha( UBI_Options["alpha"] );

        -- Redraw Options frame
        UberInventory_SetOptions( frame );
    end;

-- Make Settings frame known to the system
    function UberInventory_AddCategory( frame )
        -- Set panel name
        frame.name = UBI_NAME;

        -- Set okay function
        frame.okay = function( self )
            UberInventory_SaveOptions( self );
            UberInventory_UpdateUI();
        end;

        -- Set cancel function
        frame.cancel = function( self )
            UberInventory_SetOptions( self );
            UberInventory_UpdateUI();
        end;

        -- set defaults function
        frame.default = function( self )
            UberInventory_SetDefaults( self );
            UberInventory_UpdateUI();
        end;

        -- Add the panel
        InterfaceOptions_AddCategory( frame );
    end;

-- Update button mask to indicate item usability (clear = usable, red = usuable, blue = uncached item)
    function UberInventory_MarkButton( button, itemid )
        -- Clear state
        button:SetAttribute( "usable", nil );

        -- Reapply color coding
        if UberInventory_UsableItem( itemid ) then
            SetItemButtonTextureVertexColor( button, 1, 1, 1 );
        else
            SetItemButtonTextureVertexColor( button, 1, 0.1, 0.1 );
        end;

        -- Item still not locally cached, mark it again
        if ( not GetItemInfo( itemid ) ) then
            button:SetAttribute( "usable", "uncached" );
            SetItemButtonTextureVertexColor( button, 0.2, 0.2, 0.8 );
        end;
    end;

-- Handle uncached items
    function UberInventory_ItemButton_OnUpdate( self, elapsed )
        -- Initialize
        local state = self:GetAttribute( "usable" );
        local total_elapsed = self:GetAttribute( "elapsed" ) + elapsed;

        -- To minimize load perform the rest only if item is uncached and concerns a visible button
        if ( state and state == "uncached" and self:IsVisible() ) then
            -- Update total elapsed time since last real call
            self:SetAttribute( "elapsed", total_elapsed );

            -- Re-mark the button and reset elapsed
            if ( total_elapsed > 0.5 ) then
                UberInventory_MarkButton( self, self:GetID() );
                self:SetAttribute( "elapsed", 0 );
            end;
        end;
    end;

-- Update cash field
    function UberInventory_UpdateCashField( frame, cash, prefix )
        -- Update cash info and resize frame
        if ( cash > 0 ) then
            frame:SetText( (prefix or "")..UberInventory_GetMoneyString( cash ).."  " );
        else
            frame:SetText( (prefix or "").."-  " );
        end;
    end;

-- Update inventory button
    function UberInventory_UpdateButton( slot, record )
        local objSlot = _G[ "UberInventoryFrameItem"..slot ];
        local button = objSlot:GetName().."ItemButton";
        local buttonObj = _G[ button ];

        -- Record is nil, item does not exist and button should be hidden
        if ( not record ) then
            objSlot:Hide();
            return;
        else
            objSlot:Show();
        end;

        -- Get itemID
        local itemid = record["itemid"];

        -- Get item prices
        local buyPrice, sellPrice = UberInventory_GetItemPrices( itemid );

        -- Get item counts
        local bagCount = record["bag_count"] or 0;
        local bankCount = record["bank_count"] or 0;
        local keyCount = record["keyring_count"] or 0;
        local mailCount = record["mailbox_count"] or 0;
        local equipCount = record["equip_count"] or 0;
        local guildCount = record["count"] or 0;
        local totalCount = record["total"] or 0;

        -- Get correct totalCount for combined inventory searches
        if ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "complete" or UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_character" or UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_guildbank" ) then
            totalCount = UBI_Track[record.itemid] or 0;
        end;

        -- Set item image
        SetItemButtonTexture( buttonObj, GetItemIcon( itemid ) );

        -- Set item name
        _G[ button.."ItemName" ]:SetText( record["name"] );
        _G[ button.."ItemName" ]:SetTextColor( ITEM_QUALITY_COLORS[record.quality].r,
                                               ITEM_QUALITY_COLORS[record.quality].g,
                                               ITEM_QUALITY_COLORS[record.quality].b );

        -- Set item count
        if ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "current" or UBI_LocationList[UBI_FILTER_LOCATIONS].type == "character" ) then
            _G[ button.."ItemCount" ]:SetFormattedText( UBI_ITEM_COUNT, totalCount, bagCount, bankCount, keyCount, mailCount, equipCount );
        else
            _G[ button.."ItemCount" ]:SetFormattedText( UBI_ITEM_COUNT_SINGLE, totalCount );
        end;
        SetItemButtonCount( buttonObj, totalCount );

        -- Set item id (for tooltip)
        buttonObj:SetID( itemid );
        buttonObj:SetAttribute( "inventoryitem", record );
        buttonObj:SetAttribute( "location", UBI_FILTER_LOCATIONS );
        buttonObj:SetAttribute( "usable", nil );
        buttonObj:SetAttribute( "elapsed", 0 );

        -- Set item buy price
        UberInventory_UpdateCashField( _G[ button.."ItemBuy" ], buyPrice or 0, UBI_ITEM_BUY );

        -- Set item sell price
        UberInventory_UpdateCashField( _G[ button.."ItemSell" ], sellPrice or 0, UBI_ITEM_SELL );

        -- Mark usability onto the itembutton
        UberInventory_MarkButton( buttonObj, itemid );
    end;

-- Display items (by page, UBI_NUM_ITEMBUTTONS items)
    function UberInventory_DisplayItems( page, str )
        -- From global to local
        local UBI_Characters = UBI_Characters;
        local UBI_Guildbanks = UBI_Guildbanks;
        local UBI_Data = UBI_Data;
        local UBI_Guildbank = UBI_Guildbank;
        local UBI_Sorted = UBI_Sorted;

        -- Create GB structure if it does not already exist
        UberInventory_Guildbank_Init();

        -- Sort and filter items
        UberInventory_SortFilter( str or "", false );

        -- Update scrollbars
        if ( UBI_DISP_COUNT ~= #UBI_Sorted ) then
            UberInventory_Scroll_Init( ceil( #UBI_Sorted / UBI_NUM_ITEMBUTTONS ) - 1 );
        end;
        UBI_DISP_COUNT = #UBI_Sorted;

        -- Calculate start item
        local index;
        local startItem = page * UBI_NUM_ITEMBUTTONS + 1;

        -- Execute code to actually add the items to the inventory frame
        for slot = 1, UBI_NUM_ITEMBUTTONS do
            -- Calculate the actual index within sorted array
            index = startItem + slot - 1;

            -- Display item
            UberInventory_UpdateButton( slot, UBI_Sorted[ index ] );
        end;

        -- Update inventory count
        UberInventoryFrameInventoryCount:SetFormattedText( UBI_INVENTORY_COUNT, #UBI_Sorted, UBI_Inventory_count );

        -- Show/Hide the moneyframe for displaying Guildbank or Alt cash
        if ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "current" ) then
            UberInventoryFrameMoneyOthersCash:SetText( UBI_EMPTY_TEXT );
        else
            -- Initialize
            local otherCash = 0;

            -- Determine the cash owned
            if ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "guildbank" ) then
               otherCash = UBI_Guildbank[UBI_LocationList[UBI_FILTER_LOCATIONS].value]["Cash"];
            elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "character" ) then
               otherCash = UBI_Data[UBI_REALM][UBI_LocationList[UBI_FILTER_LOCATIONS].value]["Money"]["Cash"];
            elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_character" ) then
                -- Traverse guildbanks
                otherCash = otherCash + UBI_Data[UBI_REALM][UBI_PLAYER]["Money"]["Cash"];

                -- Traverse characters
                for key, value in pairs( UBI_Characters ) do
                    otherCash = otherCash + UBI_Data[UBI_REALM][value]["Money"]["Cash"];
                end;
            elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "all_guildbank" ) then
                -- Traverse guildbanks
                for key, value in pairs( UBI_Guildbanks ) do
                    otherCash = otherCash + UBI_Guildbank[value]["Cash"];
                end;
            elseif ( UBI_LocationList[UBI_FILTER_LOCATIONS].type == "complete" ) then
                -- Traverse guildbanks
                otherCash = otherCash + UBI_Data[UBI_REALM][UBI_PLAYER]["Money"]["Cash"];

                -- Traverse characters
                for key, value in pairs( UBI_Characters ) do
                    otherCash = otherCash + UBI_Data[UBI_REALM][value]["Money"]["Cash"];
                end;

                -- Traverse guildbanks
                for key, value in pairs( UBI_Guildbanks ) do
                    otherCash = otherCash + UBI_Guildbank[value]["Cash"];
                end;
            end;

            -- Update the cash info for Alt/Guildbank
            UberInventory_UpdateCashField( UberInventoryFrameMoneyOthersCash, otherCash );
        end;
    end;

-- Update wallet/mail cash info on screen
    function UberInventory_WalletMailCashInfo()
        if(  UberInventoryFrame:IsVisible() ) then
            -- Update cash balance information
            UberInventory_UpdateCashField( UberInventoryFrameMoneyWalletCash, GetMoney() );
            UberInventory_UpdateCashField( UberInventoryFrameMoneyMailCash, UBI_Money["mail"] );
        end;
    end;

-- Show or hide the Main dialog
    function UberInventory_Toggle_Main()
        if(  UberInventoryFrame:IsVisible() ) then
            UberInventoryFrame:Hide();
        else
            -- Show the frame
            UberInventoryFrame:Show();
            UberInventoryFrame:SetBackdropColor( 0, 0, 0, 1 );

            -- Update wallet/mail cash information
            UberInventory_WalletMailCashInfo();

            -- Force refresh
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT or "" );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;
    end;

-- Set/Remove hightlight for container/item
    function UberInventory_Highlighter( item, state )
        -- Set hightlight from container/item
        function HighlightOn( frame )
            if ( frame:GetName() ~= "KeyRingButton" ) then
                SetItemButtonTextureVertexColor( frame, .6, .6, .6, 1 );
            end;
            frame:GetNormalTexture():SetVertexColor( .6, .6, .6, 1 );
            frame:LockHighlight();
            tinsert( UBI_Highlights, frame );
        end;

        -- Remove hightlight from container/item
        function HighlightOff( frame )
            if ( frame:GetName() ~= "KeyRingButton" ) then
                SetItemButtonTextureVertexColor( frame, 1, 1, 1 );
            end;
            frame:GetNormalTexture():SetVertexColor( 1, 1, 1 );
            frame:UnlockHighlight();
        end;

        -- Exit if highlighting has been turned off
        if ( not UBI_Options["show_highlight"] ) then
            return;
        end;

        -- Set or remove highlights
        if ( state == "off" ) then
            -- Remove all previous highlights
            for id, frame in pairs( UBI_Highlights ) do
                HighlightOff( frame );
            end;
            wipe( UBI_Highlights );
        else
            -- Initialize
            local bagid, slotid, itemid, itemlink, index, container;

            -- Travese all containers and slots (bag, keyring, bank) to locate items
            for bagid = -2, 11, 1 do
                index = IsBagOpen( bagid );
                for slotid = 1, GetContainerNumSlots( bagid ) do
                    itemlink = GetContainerItemLink( bagid, slotid );
                    if ( itemlink ) then
                        _, itemid = strsplit( ":", itemlink );
                        if ( tonumber(itemid) == item and UI_CONTAINER_OBJECTS[bagid]:IsShown() ) then
                            -- Highlight container
                            if ( bagid ~= -1 ) then
                                HighlightOn( UI_CONTAINER_OBJECTS[bagid] );
                            end;

                            -- Highlight item
                            if ( index ) then
                                if ( bagid == -1 ) then
                                    container = "BankFrameItem"..slotid;
                                elseif ( bagid == -2 ) then
                                    slotid = GetKeyRingSize() - slotid + 1;
                                    container = "ContainerFrame"..index.."Item"..slotid;
                                else
                                    slotid = GetContainerNumSlots( bagid ) - slotid + 1;
                                    container = "ContainerFrame"..index.."Item"..slotid;
                                end;
                                HighlightOn( getglobal( container ) );
                            end;
                        end;
                    end;
                end;
            end;

            -- Travese all gb slots to locate item
            if ( UBI_GUILDBANK_OPEN ) then
                local gbTab, gbSlot;
                for gbTab = 1, GetNumGuildBankTabs() do
                    for gbSlot = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
                        itemlink = GetGuildBankItemLink( gbTab, gbSlot );
                        if ( itemlink ) then
                            _, itemid = strsplit( ":", itemlink );
                            if ( tonumber(itemid) == item ) then
                                -- Highlight tab
                                container = "GuildBankTab"..gbTab.."Button";
                                HighlightOn( getglobal( container ) );

                                -- Highlight item
                                if ( GetCurrentGuildBankTab() == gbTab ) then
                                    index = mod( gbSlot, NUM_SLOTS_PER_GUILDBANK_GROUP );
                                    if ( index == 0 ) then
                                        index = NUM_SLOTS_PER_GUILDBANK_GROUP;
                                    end
                                    column = ceil( ( gbSlot - 0.5 ) / NUM_SLOTS_PER_GUILDBANK_GROUP );
                                    container = "GuildBankColumn"..column.."Button"..index;
                                    HighlightOn( getglobal( container ) );
                                end;
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;

-- Perform item search
    function UberInventory_Search( str )
        -- From global to local
        local itemText, msg, count = "", "", 0;

        -- Display search criteria
        UberInventory_Message( UBI_ITEM_SEARCH:format( str ) );

        -- Search and display items found
        local itemCounts = { 0, 0, 0, 0, 0, 0 };
        for key, value in pairs( UBI_Items ) do
            if ( ( strfind( strlower( value.name ) , strlower( str ) ) or 0 ) > 0 ) then
                -- Get item count
                itemCounts = { value.total,
                               value.bag_count or 0,
                               value.bank_count or 0,
                               value.keyring_count or 0,
                               value.mailbox_count or 0,
                               value.equip_count or 0 };

                -- Build search result
                itemText = "|T"..GetItemIcon( value.itemid )..":0|t "..UberInventory_GetLink( value.itemid );
                msg = "  "..itemText.." "..UBI_ITEM_COUNT_SINGLE:format( itemCounts[1] ).." (";
                for loc = 2, #itemCounts do
                    if ( itemCounts[loc] > 0 ) then
                        msg = msg.."|T"..UBI_LOCATION_TEXTURE[loc-1]..":0|t "..UBI_LOCATIONS[loc-1].." = "..itemCounts[loc].." / ";
                    end;
                end;

                -- Display result
                count = count + 1;
                UberInventory_Message( msg:sub( 1, msg:len()-3 )..")" );
            end;
        end;

        -- Display end of search
        if ( count == 0 ) then
           UberInventory_Message( UBI_ITEM_SEARCH_NONE );
        else
           UberInventory_Message( UBI_ITEM_SEARCH_DONE );
        end;
    end;

-- Remove character or guildbank data
    function UberInventory_RemoveData( data )
        -- split of command and name (guildbank names can contain spaces, unlike character names)
        local command, _ = strsplit( " ", data ):lower();
        local name = data:sub( command:len()+2 );

        -- Remove the data
        if ( command == "remchar"  and name ~= "Guildbank" ) then
            -- Do not remove data for the current connected character
            if ( name == UBI_PLAYER ) then
                UberInventory_Message( UBI_REM_NOTALLOWED );
                return;
            end;

            -- Check if character has data stored and delete it if so
            if ( UBI_Data[UBI_REALM][name] ) then
                StaticPopupDialogs["UBI_CONFIRM_DELETE"].OnAccept = function()
                    -- Remove data
                    UBI_Data[UBI_REALM][name] = nil;

                    -- Display message
                    UberInventory_Message( UBI_REM_DONE:format( name ) );

                    -- Rebuild list of alts
                    UberInventory_GetAlts();

                    -- Reset filters
                    UberInventory_ResetFilters();
                end;

                -- Ask for remove confirmation
                StaticPopup_Show( "UBI_CONFIRM_DELETE", UBI_REM_CHARACTER:format( name ) );
            else
                UberInventory_Message( UBI_REM_CHARNOTFOUND:format( name ) );
            end;
        elseif ( command == "remguild" ) then
            -- Check if guildbank has data stored and delete it if so
            if ( UBI_Guildbank[name] ) then
                StaticPopupDialogs["UBI_CONFIRM_DELETE"].OnAccept = function()
                    -- Remove data
                    UBI_Guildbank[name] = nil;

                    -- Display message
                    UberInventory_Message( UBI_REM_DONE:format( name ) );

                    -- Rebuild list of guildbanks
                    UberInventory_GetGuildbanks();

                    -- Reset filters
                    UberInventory_ResetFilters();
                end;

                -- Ask for remove confirmation
                StaticPopup_Show( "UBI_CONFIRM_DELETE", UBI_REM_GUILDBANK:format( name ) );
            else
                UberInventory_Message( UBI_REM_GUILDNOTFOUND:format( name ) );
            end;
        end;
    end;

-- Show help message
    function UberInventory_ShowHelp()
        for key, value in pairs( UBI_HELP ) do
            UberInventory_Message( value );
        end;
    end;

-- Handle slash commands
    function UberInventory_SlashHandler( msg )
        -- arguments should be handled case-insensitve
        local orgData = msg;
        msg = strlower( msg );

        -- Handle each individual argument
        if ( msg == "" ) then
            -- Show main dialog
            UberInventory_Toggle_Main();
        elseif ( msg:find( "remchar" ) or msg:find( "remguild" ) ) then
            -- Remove character or guild data
            UberInventory_RemoveData( orgData );
        elseif ( msg == "help" ) or ( msg == "?" ) then
            -- Show help info
            UberInventory_ShowHelp();
        elseif ( msg == "resetpos" ) then
            UberInventoryFrame:ClearAllPoints();
            UberInventoryFrame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
        else
            -- Perform item search
            UberInventory_Search( msg or "" );
        end;
    end;
